import json
import socket as socket_module
import ssl
from pathlib import Path

from doip_sdk.model.response import Response


def write_json_segment(socket: socket_module.socket, message: dict, encoding: str = 'utf-8'):
    """
    Write a JSON segment to the provided socket. The segment ends with a line that only contains a ``#``.

    Parameters
    ----------
    socket : :py:class:`socket <python:socket.socket>`
        The socket to write the JSON segment to.
    message : :py:class:`dict <python:dict>`
        The JSON segment to write.
    encoding : :py:class:`str <python:str>`, default='utf-8'
        The encoding which will be used to encode the JSON segment.
    """
    message_byte = bytes(json.dumps(message), encoding)
    socket.sendall(message_byte + b'\n#\n')


def write_file_segment(socket: socket_module.socket, file: Path, chunk_size: int = 1024 * 1024):
    """
    Write a byte segment to the provided socket.

    Parameters
    ----------
    socket : :py:class:`socket <python:socket.socket>`
        The socket to write the byte segment to.
    file : :py:class:`Path <python:pathlib.Path>`
        The file that will be sent.
    chunk_size : :py:class:`int <python:int>`, default=1 MB
        The size of each chunk in bytes. Default to 1 MB.
    """
    socket.sendall(b'@\n')
    with file.open('rb') as f:
        while bytes_read := f.read(chunk_size):
            chunk_size_byte_string = bytes(f'{len(bytes_read)}\n', encoding='utf-8')
            socket.sendall(chunk_size_byte_string)
            socket.sendall(bytes_read + b'\n')
    socket.sendall(b'#\n')


def write_empty_segment(socket: socket_module.socket):
    """
    Write an empty segment to the provided socket. An empty segment is just a line with only a ``#`` sign.

    Parameters
    ----------
    socket : :py:class:`socket <python:socket.socket>`
        The socket to write the empty segment to.
    """
    socket.sendall(b'#\n')


def send_request(host: str, port: int, payload: list[dict | Path], verify_host: bool = False,
                 stream: bool = False) -> Response:
    """
    Send a DOIP request to a server. If the response from the server is huge, the streaming response mode can be
    activated by setting ``stream`` to ``True``. The response need to be read differently based on the ``stream``
    setting.

    Parameters
    ----------
    host : :py:class:`str <python:str>`
        The address of the server.
    port : :py:class:`int <python:int>`
        The port of the server.
    payload: :py:class:`list[dict | Path] <python:list>`
        A list of segments to be sent. Each element can be a :py:class:`dictionary <python:dict>` for a JSON segment, or
        a :py:class:`Path <python:pathlib.Path>` for a byte segment.
    verify_host : :py:class:`bool <python:bool>`, default=``False``
        Should the server's certificate be checked or not.
    stream : :py:class:`bool <python:bool>`, default=``False``
        Should the response be streamed or not.

    Returns
    -------
    :py:class:`~doip_sdk.model.response.Response`
        The response from the server is stored in this object. The actual data can be access via the ``content``
        attribute. If the streaming mode is off, ``content`` is a
        :py:class:`list[bytearray] <python:list>`, where each element is a segment. Otherwise, ``content`` is an
        :py:class:`Iterator[bytearray] <python:collections.abc.Iterator>`.
    """
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    if not verify_host:
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE

    secured_socket = None
    try:
        with socket_module.create_connection((host, port)) as socket:
            secured_socket = context.wrap_socket(socket, server_hostname=host)

            # Send request
            for item in payload:
                if isinstance(item, dict):
                    write_json_segment(socket=secured_socket, message=item)
                elif isinstance(item, Path):
                    write_file_segment(socket=secured_socket, file=item)
            write_empty_segment(socket=secured_socket)

            # Process response
            return Response(socket=secured_socket, stream=stream)
    except Exception as e:
        # Do not close the socket in the finally, because it should be close in the Response instead
        if secured_socket:
            secured_socket.close()
        raise e



import ssl
from pathlib import Path
from socketserver import ThreadingTCPServer

from doip_sdk.base.handler import DOIPHandler
from doip_sdk.selfsigned import generate_selfsigned_cert


class DOIPServer(ThreadingTCPServer):
    """
    This is the server which is used to handle DOIP requests.

    Parameters
    ----------
    host : :py:class:`str <python:str>`
        The address of the server.
    port : :py:class:`int <python:int>`
        The port of the server.
    handler : :py:class:`~doip_sdk.base.handler.DOIPHandler`
        The handler used to handle requests. This handler must inherit the
        :py:class:`~doip_sdk.base.handler.DOIPHandler`.
    key_cert_files : :py:class:`tuple[str, str] <python:tuple>`, default=None
        A tuple, where the first element is the private key and the second is the certificate. If this parameter is not
        provided, the server will generate a self-signed certificate and use it.
    bind_and_activate : :py:class:`bool <python:bool>`, default=True
        Should the server be activated immediately in the constructor. One might need to set it to ``False`` when they
        intend to create a custom server based on this ``DOIPServer``.

    Examples
    --------
    >>> HOST, PORT = 'localhost', 9999
    >>> with DOIPServer(HOST, PORT, ExampleHandler) as server:
    >>>     server.start()
    """

    def __init__(self, host: str, port: int, handler: DOIPHandler, key_cert_files: tuple[str, str] | None = None,
                 bind_and_activate=True):
        super(DOIPServer, self).__init__((host, port), handler, bind_and_activate=False)

        # Auto generate a self-signed certificate, if not provided
        if key_cert_files:
            key_file, cert_file = key_cert_files
        else:
            key, cert = generate_selfsigned_cert(hostname=host)

            key_file = Path('ssl/key.pem')
            cert_file = Path('ssl/cert.pem')
            key_file.parent.mkdir(parents=True, exist_ok=True)

            with key_file.open('wb+') as f:
                f.write(key)
            with cert_file.open('wb+') as f:
                f.write(cert)

        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain(certfile=cert_file, keyfile=key_file)
        self.socket = context.wrap_socket(self.socket, server_side=True)

        # Keep the paths so that we can delete them later
        self.key_file = key_file
        self.cert_file = cert_file

        if bind_and_activate:
            self.server_bind()
            self.server_activate()

    def server_close(self):
        """
        Call this method to clean up resources before shutting down the server. It is recommended to use the server with
        the ``with`` statement to avoid calling this method manually.
        """
        super().server_close()

        # Remove self-signed certificate
        self.key_file.unlink(missing_ok=True)
        self.cert_file.unlink(missing_ok=True)
        self.key_file.parent.rmdir()

    def start(self):
        """Call this method to start the server. To stop the server, simply press :kbd:`ctrl` + :kbd:`c`."""
        try:
            self.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            print('Shutting down the server')



__all__ = [
    'DOIPHandler',
    'Operation',
    'ResponseStatus',
    'ServerResponse',
    'Response',
    'write_json_segment',
    'write_empty_segment',
    'write_file_segment',
    'send_request',
    'SocketReader',
    'DOIPServer'
]

from doip_sdk.base.handler import DOIPHandler
from doip_sdk.base.server import DOIPServer
from doip_sdk.constant import Operation, ResponseStatus
from doip_sdk.model.response import ServerResponse, Response
from doip_sdk.socket_reader import SocketReader
from doip_sdk.utils import write_json_segment, write_empty_segment, write_file_segment, send_request



import socket as socket_module
from collections.abc import Iterator
from typing import Any

from pydantic import BaseModel, ConfigDict

from doip_sdk.socket_reader import SocketReader
from doip_sdk.constant import ResponseStatus


class ServerResponse(BaseModel):
    model_config = ConfigDict(
        # So that when we serialize the model, we get the string value of the enum instead of the enum object
        use_enum_values=True
    )

    requestId: str | None = None
    status: ResponseStatus
    attributes: list = None
    output: Any | None = None


class Response:
    """
    This class represents the response from the server. The payload of the response can be accessed via the ``content``
    attribute.

    Parameters
    ----------
    socket : :py:class:`socket <python:socket.socket>`
        The socket used to read the response.
    stream : :py:class:`bool <python:bool>`, default=``False``
        Should the response be streamed or not.

    Attributes
    ----------
    socket : :py:class:`socket <python:socket.socket>`
        The socket used to read the response.
    reader : :py:class:`~doip_sdk.socket_reader.SocketReader`
        A :py:class:`~doip_sdk.socket_reader.SocketReader`, which is used to parse the response.
    content : :py:class:`list[bytearray] <python:list>` | :py:class:`Iterator[bytearray] <python:collections.abc.Iterator>`
        The actual content of the response. If the streaming mode is off, ``content`` is a
        :py:class:`list[bytearray] <python:list>`, where each element is a segment. Otherwise, ``content`` is an
        :py:class:`Iterator[bytearray] <python:collections.abc.Iterator>`, where each element can either be:

        1. a JSON segment, or
        2. an ``@`` sign, which indicates next elements are chunks in a byte segment, or
        3. a chunk in a byte segment, or
        4. an ``#`` sign, which indicates the end of a byte segment.
    """

    socket: socket_module
    reader: SocketReader
    content: list[bytearray] | Iterator[bytearray]

    def __init__(self, socket: socket_module.socket, stream: bool = False):
        self.socket = socket
        self.reader = SocketReader(socket=socket)

        if stream:
            self.content = self.reader.get_chunks()
        else:
            try:
                self.content = self._read_response_as_list()
            finally:
                self.close()

    def _read_response_as_list(self) -> list[bytearray]:
        segments = []
        chunks = self.reader.get_chunks()
        while True:
            segment = next(chunks, None)
            if not segment:
                break
            if segment != b'@':
                segments.append(segment)
            else:
                file = bytearray()
                while True:
                    file_chunk = next(chunks)
                    if file_chunk == b'#':
                        break
                    file.extend(file_chunk)
                segments.append(file)
        return segments

    def close(self):
        """
        Must be called to close the socket when ``stream=True``. It is recommended to use the ``with`` statement when
        reading the response stream so that one do not have to call this method manually.
        """
        if self.socket:
            self.socket.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
